'''A socket connection to a Zimi Controller.'''

import asyncio
from codecs import StreamReader, StreamWriter
import json
from json.decoder import JSONDecodeError
import logging
import queue
import socket

from zcc.constants import LEVEL_BY_VERBOSITY


class ControlPointSocket:
    '''A TCP/IP socket that includes recvall method with timeout'''

    def __init__(self, host: str, port: int, timeout: int = None, verbosity: int = 0):
        '''Create new TCP/IP socket to host and port with optional timeout.
        Update objects are made available via a FIFO queue.'''

        self.host = host
        self.port = port

        self._observers = []

        self.listen_task: asyncio.Task = None

        self.listen_queu = queue.Queue()
        self.send_queu = queue.Queue()

        self.loop = asyncio.get_event_loop()

        self.logger = logging.getLogger('ControlPointSocket')
        self.logger.setLevel(LEVEL_BY_VERBOSITY[verbosity])

        self.sock: socket.socket = None
        self.reader: StreamReader = None
        self.writer: StreamWriter = None

        self.timeout = timeout

    async def connect(self):
        '''Connect to an OS socket and begin listening.'''

        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        if self.sock and self.timeout:
            self.sock.settimeout(self.timeout)

        try:
            self.sock.connect((self.host, self.port))
            self.reader, self.writer = await asyncio.open_connection(sock=self.sock)
            self.listen_task = asyncio.create_task(self.__listen())
        except asyncio.CancelledError as error:
            self.logger.error("Connection cancelled")
            raise error
        except ConnectionRefusedError as error:
            self.logger.error(
                'Connection refused %s error from %s:%s',
                error, self.host, str(self.port))
            raise error
        except socket.error as error:
            self.logger.error(
                'Connection received socket %s error from %s:%s',
                error, self.host, str(self.port))
            raise error

    async def __listen(self):
        '''Worker thread to monitor state change events and update objects.
        Sends a notify to observers when new items are put in the queue OR the
        underlying socket is closed.'''

        while True:
            try:
                line = await asyncio.wait_for(self.reader.readline(), timeout=1)
                if line == b'':
                    continue
                response = json.loads(line).get('response')
                if response:
                    self.listen_queu.put(response)
                    for obs in self._observers:
                        await obs.notify(self)
                    self.send_queu.get(block=False)
            except asyncio.CancelledError:
                self.logger.error("Listening connection cancelled")
                break
            except asyncio.TimeoutError:
                if self.sock and self.sock.fileno() != -1 and self.reader:
                    pass
                else:
                    self.logger.error(
                        'Listening connection socket failure after timeout')
                    break
            except AttributeError:
                self.logger.error(
                    "Listening connection cancelled during timeout")
                break
            except (ConnectionResetError, ConnectionAbortedError) as error:
                self.logger.error(
                    "Llstening connection reset error %s", error)
                break
            except JSONDecodeError as error:
                self.logger.error(
                    "Listening connection JSON decode error %s:\n%s", error, line)
                break
            except queue.Empty:
                pass

        self.logger.debug("Listening connection socket closed")

        self.listen_task = None  # Remove reference before closing
        self.close()

        for obs in self._observers:
            await obs.notify(self)

    def close(self) -> None:
        '''Close socket'''
        if self.listen_task:
            self.listen_task.cancel()
        if self.sock:
            self.sock.close()
        self.reader = None
        self.writer = None
        self.sock = None

    def get(self):
        '''Fetch an item from the queue.  Returns None if the queue is empty.
        Notify for an empty queue should have only been sent if the socket has
        been closed and an extra notify() was sent.'''

        try:
            response = self.listen_queu.get(block=False)
        except queue.Empty:
            response = None

        return response

    async def sendall(self, message: str, response_expected: bool = True) -> bool:
        '''Send all bytes of string to socket end-point.   If the socket is closed
        send a None notify to prompt re-connect.  If response_expected is set
        the message is saved in a queu incase it needs to be retrieved to send again.'''

        if response_expected:
            self.send_queu.put(message)

        if self.sock and self.sock.fileno() != -1 and self.writer and not self.writer.is_closing():
            self.logger.debug('sendall()\n%s', message)
            self.writer.write(message.encode())
            await self.writer.drain()
            return True
        else:
            self.logger.error('Send failed with no socket:\n%s', message)
            self.close()
            for obs in self._observers:
                await obs.notify(self)
            return False

    def subscribe(self, observer):
        '''Subscribe an observer object for state changes.
        Observer object must include an async notify(self, observable, *args, **kwargs) method.'''
        self._observers.append(observer)

    def unsubscribe(self, observer):
        '''Unsubscribe an observer object.'''
        try:
            self._observers.remove(observer)
        except ValueError as error:
            self.logger.debug(
                "Unsubscribe failed with value error: %s for %s", error, observer)

    def unsent(self):
        '''Fetch an item from the unsent queue.  Returns None if the queue is empty.'''

        try:
            response = self.send_queu.get(block=False)
        except queue.Empty:
            response = None

        return response



'''ZCC Trace Class'''

import imp
import logging
import os
import pkgutil
import tracemalloc

from zcc.constants import NAME
from zcc.watchdog import ControlPointWatchdog

_logger = logging.getLogger('ControlPointTrace')


def _package_files(package):
    '''Yeild files in moddule.
    Ref https://stackoverflow.com/a/36764029'''

    _, pathname, _ = imp.find_module(package)
    for dirpath, _, filenames in os.walk(pathname):
        for filename in filenames:
            if os.path.splitext(filename)[1] == ".py":
                yield os.path.join(dirpath, filename)


class ControlPointTrace:
    '''A class to show zcc statistics'''

    def __init__(self, timer: int):

        self.logger = logging.getLogger('ControlPointTrace')

        tracemalloc.start()

        self.statistics_timer = ControlPointWatchdog(
            timer, self.print_statistics)

        self.statistics_timer.start()

    async def print_statistics(self):
        '''Print statistics and re-start timer'''
        print(
            'Printing module memory statistics:\n' + self._statistics)
        self.statistics_timer.reset()

    @property
    def _statistics(self) -> str:
        '''Return statistics of the ZCC module.'''

        snapshot = tracemalloc.take_snapshot()

        filename_filters = []

        for filename in _package_files('zcc'):
            filename_filters.append(tracemalloc.Filter(
                inclusive=True, filename_pattern=filename))

        snapshot = snapshot.filter_traces(filename_filters)

        summary_count = 0
        summary_bytes = 0

        line_details = ''

        for stat in snapshot.statistics('lineno'):
            summary_count += stat.count
            summary_bytes += stat.size
            line_details += str(stat) + '\n'

        summary_line = 'Module has %d objects and totalling %d bytes with lines:\n' % (
            summary_count, summary_bytes)

        return summary_line + line_details



'''ControlPointDevice Class.'''
from __future__ import annotations

from pprint import pformat

from zcc.errors import ControlPointError


class ControlPointDevice:
    '''Represents any ControlPointDevice.'''

    def __init__(self, controller, identifier: str):
        '''Create a new device associated with a controller and identifed by identifier'''
        self.controller = controller
        self.identifier = identifier
        self.actions = {}
        self.properties = {}
        self.states = {}

        self._target_percentage = None
        self._observers = []

    def __str__(self):
        return pformat(vars(self))

    async def __action(self, action: str, params: object = None) -> bool:
        try:
            if self.actions['actions'][action]:
                await self.controller.set(self.identifier, action, params=params)
        except Exception as exception_info:
            raise ControlPointError(
                action + " is not supported for " + self.identifier) from exception_info

    @property
    def __actions(self) -> str:
        '''Gets a descriptive string of what actions are supported'''
        description = '{ '
        try:
            if self.actions['actions'].get("TurnOn", None):
                description += "TurnOn TurnOff "
            if self.actions['actions'].get("SetBrightness", None):
                description += "SetBrightness "
            if self.actions['actions'].get("OpenDoor", None):
                description += "OpenDoor CloseDoor "
            if self.actions['actions'].get("OpenToPercentage", None):
                description += "OpenToPercentage "
            if self.actions['actions'].get("SetFanSpeed", None):
                description += "SetFanSpeed "
        except KeyError:
            pass
        description += '}'
        if description == '{ }':
            description = ''
        return description

    @property
    def __states(self) -> str:
        '''Gets a descriptive string of the device state'''
        description = ''
        try:
            key = list(self.states['controlState'].keys())[0]
            state = self.states['controlState'][key]
            description += "On" if state.get('isOn', False) else "Off"
            brightness = state.get('brightness', None)
            if brightness:
                description += '/' + str(brightness)
            fan_speed = state.get('fanspeed', None)
            if fan_speed:
                description += '/' + str(fan_speed)
        except IndexError:
            pass
        except KeyError:
            pass
        return description

    @property
    def battery_level(self) -> int | None:
        '''Return the battery level of an attached sensor.'''
        try:
            state = self.states['controlState']['sensor']
            return state.get('batterylevel', 0)
        except KeyError:
            return None

    @property
    def brightness(self) -> int | None:
        '''Returns brightness from 0 to 100 or None.'''
        try:
            key = list(self.states['controlState'].keys())[0]
            state = self.states['controlState'][key]
            brightness = state.get('brightness', None)
            if brightness:
                return brightness
        except KeyError:
            return False

    async def close_door(self):
        '''CloseDoor if the action is supported'''
        self._target_percentage = 0
        await self.__action("CloseDoor")

    def describe(self) -> str:
        '''Returns a description of a device'''
        return '%-40s %-40.40s %-8.8s %-8s %s\n' % \
            (self.identifier,
             self.location,
             self.type,
             self.__states,
             self.__actions)

    @property
    def door_temp(self) -> int | None:
        '''Return the external temperature of an attached sensor.'''
        try:
            state = self.states['controlState']['sensor']
            return state.get('doortemp', 0)
        except KeyError:
            return None

    async def fade(self, brightness, timeperiod):
        '''SetBrightness if the action is supported'''
        await self.__action("SetBrightness", params={
            "brightness": int(brightness),
            "timeperiod": int(timeperiod)})

    @ property
    def fanspeed(self) -> int | None:
        '''Returns fanspeed from 0 to 7 or None.'''
        try:
            key = list(self.states['controlState'].keys())[0]
            state = self.states['controlState'][key]
            fanspeed = state.get('fanspeed', None)
            if fanspeed:
                return fanspeed
        except KeyError:
            return False

    @property
    def garage_temp(self) -> int | None:
        '''Return the internal garage temperature of an attached sensor.'''
        try:
            state = self.states['controlState']['sensor']
            return state.get('garagetemp', 0)
        except KeyError:
            return None

    @property
    def garage_humidity(self) -> int | None:
        '''Return the internal garage humidity of an attached sensor.'''
        try:
            state = self.states['controlState']['sensor']
            return state.get('garagehumidity', 0)
        except KeyError:
            return None

    @ property
    def is_closing(self) -> bool:
        '''Returns True if door is closing.'''
        if self.percentage:
            if self.percentage > self._target_percentage:
                return True
            else:
                return False
        return False

    @ property
    def is_closed(self) -> bool:
        '''Returns True if door is closed.'''

        if self.percentage:
            if self.percentage == 0:
                return True
            else:
                return False
        return True

    @ property
    def is_connected(self) -> bool:
        '''Returns True if connected is on.
        When a device has been disconnected from the mesh it show False.'''

        try:
            return self.states['isConnected']
        except KeyError:
            return False

    @ property
    def is_off(self) -> bool:
        '''Returns True if status is off.'''

        return not self.is_on

    @ property
    def is_on(self) -> bool:
        '''Returns True if status is on.'''

        try:
            key = list(self.states['controlState'].keys())[0]
            state = self.states['controlState'][key]
            return state.get('isOn', False)
        except KeyError:
            return False

    @ property
    def is_opening(self) -> bool:
        '''Returns True if door is opening.'''

        if self.percentage:
            if self.percentage < self._target_percentage:
                return True
            else:
                return False
        return False

    @ property
    def is_open(self) -> bool:
        '''Returns True if door is open.'''

        if self.percentage:
            if self.percentage > 0:
                return True
            else:
                return False
        return False

    @ property
    def location(self) -> str:
        '''Gets a descriptive string of the device location'''
        description = self.properties.get('name', '-') + '/'
        description += self.properties.get('roomName', '-')
        return description

    @ property
    def name(self) -> str:
        '''Gets a descriptive string of the device name'''
        return self.properties.get('name', '-')

    def notify_observers(self):
        '''Notify all observers that a state change has occurred.'''
        for obs in self._observers:
            obs.notify(self)

    async def open_door(self):
        '''OpenDoor if the action is supported'''
        self._target_percentage = 100
        await self.__action("OpenDoor")

    async def open_to_percentage(self, percentage):
        '''OpenToPercentage if the action is supported'''
        self._target_percentage = percentage
        await self.__action("OpenToPercentage", params={
            "openpercentage": int(percentage)})

    @ property
    def percentage(self) -> int | None:
        '''Return the open to percentage'''
        try:
            key = list(self.states['controlState'].keys())[0]
            state = self.states['controlState'][key]
            return state.get('openpercentage', 0)
        except KeyError:
            return None

    def print_description(self):
        '''Print device description.'''
        print(self.describe())

    @ property
    def room(self) -> str:
        '''Gets a descriptive string of the device room'''
        return self.properties.get('roomName', '-')

    async def set_brightness(self, brightness):
        '''SetBrightness if the action is supported'''
        await self.__action("SetBrightness", params={"brightness": int(brightness)})

    async def set_fanspeed(self, fanspeed):
        '''SetFanSpeed if the action is supported'''
        await self.__action("SetFanSpeed", params={"fanspeed": int(fanspeed)})

    def subscribe(self, observer):
        '''Subscribe an observer object for state changes.
        Observer object must include notify(self, observable, *args, **kwargs) method.'''
        self._observers.append(observer)

    async def turn_on(self):
        '''TurnOn the device if the action is supported'''
        await self.__action("TurnOn")

    async def turn_off(self):
        '''TurnOff the device if the action is supported'''
        await self.__action("TurnOff")

    @ property
    def type(self) -> str:
        '''Gets a descriptive string of the device type'''
        return self.properties.get('controlPointType', '-')

    def unsubscribe(self, observer):
        '''Unsubscribe an observer object.'''
        self._observers.remove(observer)



'''ZCC Discovery Service Class'''
from __future__ import annotations

import asyncio
import json
from json.decoder import JSONDecodeError
import logging
import socket
from typing import Tuple

from zcc.constants import LEVEL_BY_VERBOSITY
from zcc.description import ControlPointDescription
from zcc.errors import ControlPointError
from zcc.protocol import ControlPointProtocol


class ControlPointDiscoveryProtocol(asyncio.DatagramProtocol):
    '''Listens for ZCC announcements on the defined UDP port.'''

    def __init__(self, discovery_complete: asyncio.Future, discovery_result: object) -> None:
        super().__init__()
        self.discovery_complete: asyncio.Future = discovery_complete
        self.discovery_result = discovery_result
        self.transport: asyncio.transports.DatagramTransport = None
        self.logger = logging.getLogger('ControlPointDiscoveryService')

    def connection_lost(self, exc) -> None:
        self.transport.close()
        return super().connection_lost(exc)

    def connection_made(self, transport: asyncio.transports.DatagramTransport) -> None:
        self.transport = transport
        return super().connection_made(transport)

    def datagram_received(self, data: bytes, addr: Tuple[str, int]) -> None:

        data = str(data.decode('UTF-8'))
        self.logger.debug(
            'datagram_received() from %s\n%s', str(addr), data)

        lines = data.split('\n')
        for line in lines:
            try:
                response = json.loads(line)
                if response:
                    self.discovery_result.brand = response['brand']
                    self.discovery_result.product = response['product']
                    self.discovery_result.mac = response['mac']
                    self.discovery_result.host = addr[0]
                    self.discovery_result.port = response['tcp']
                    self.discovery_result.available_tcps = response['availableTcps']
                    if api_version := response.get('apiVersion'):
                        self.discovery_result.api_version = api_version
                    if firmware_version := response.get('firmwareVersion'):
                        self.discovery_result.firmware_version = firmware_version
                    self.discovery_complete.set_result(True)
            except JSONDecodeError:
                break

        return super().datagram_received(data, addr)


class ControlPointDiscoveryService:
    '''Provides a ZCC discovery service to discover ZIMI controllers on the local LAN.'''

    def __init__(self, verbosity: int = 0):

        self.logger = logging.getLogger('ControlPointDiscoveryService')
        if verbosity > 2:
            verbosity = 2
        self.logger.setLevel(LEVEL_BY_VERBOSITY[verbosity])

        self.loop = asyncio.get_event_loop()
        self.discovery_complete = self.loop.create_future()
        self.discovery_result = ControlPointDescription()

    async def discover(self) -> ControlPointDescription:
        '''Discover local ZIMI controllers on LAN and return (host,port).'''

        transport, _ = await self.loop.create_datagram_endpoint(
            lambda: ControlPointDiscoveryProtocol(
                self.discovery_complete, self.discovery_result),
            local_addr=('0.0.0.0', ControlPointProtocol.UDP_RECV_PORT))

        send_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        send_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        send_socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)

        server_address = ('255.255.255.255',
                          ControlPointProtocol.UDP_SEND_PORT)
        message = ControlPointProtocol.discover()

        send_socket.sendto(message.encode(), server_address)
        self.logger.info('Sending discovery message on local network')

        try:
            await asyncio.wait_for(self.discovery_complete, timeout=10)
            self.logger.info("Success - discovered ZIMI controller at %s port %d\n%s",
                             self.discovery_result.host,
                             self.discovery_result.port,
                             self.discovery_result)
            transport.close()
            return self.discovery_result
        except asyncio.exceptions.TimeoutError as error:
            raise ControlPointError(
                "Failure - Unable to discover ZCC by UDP broadcast.") from error



"""Constants for ZCC."""

import logging

LEVEL_BY_VERBOSITY = [logging.WARNING, logging.INFO, logging.DEBUG]

APP_NAME = "HomeControl"
APP_ID = "ZvWqWJQB"
APP_TOKEN = "3422ecbb-cf6a-404c-b3dc-2023dd213535"

NAME = "zcc-helper"
VERSION = "3.2.2rc3"



'''ZCC Zimi Controller Description Class'''
from __future__ import annotations

from dataclasses import dataclass


@dataclass
class ControlPointDescription:
    '''Data class to store ControlPoint description.'''

    brand: str = None
    product: str = None
    mac: str = None
    host: str = None
    port: int = None
    available_tcps: int = None
    api_version: str = 'n/a'
    firmware_version: str = 'n/a'



'''ZCC Support Library to support ZIMI based home network elements'''
import logging
import sys

from .controller import ControlPoint
from .description import ControlPointDescription
from .discovery import ControlPointDiscoveryService
from .errors import ControlPointError

logging.basicConfig(stream=sys.stderr)



'''Static methods for the Zimi Controller protocol.'''

import json

from .constants import APP_ID, APP_TOKEN


class ControlPointProtocol:
    '''Static methods for the Zimi Controller protocol.'''

    AUTH_APP_FAILED = 'auth_app_failed'
    AUTH_APP_SUCCESS = 'auth_app_success'
    AUTH_TIMEOUT = 30
    START_SESSION_FAILED = 'start_session_failed'
    START_SESSION_SUCCESS = 'start_session_success'
    START_SESSION_TIMEOUT = 30
    DEVICE_GET_TIMEOUT = 10
    CONTROLPOINT_PROPERTIES = 'controlpoint_properties'
    CONTROLPOINT_STATES = 'controlpoint_states'
    CONTROLPOINT_ACTIONS = 'controlpoint_actions'
    CONTROLPOINT_SETACTIONS = 'controlpoint_setactions'
    CONTROLPOINT_STATES_EVENTS = 'controlpoint_states_events'
    ZCC_STATUS = 'zcc_status'
    
    SUBSCRIBE_TIMEOUT = 5
    GATEWAY_PROPERTIES = 'gateway_properties'
    GATEWAY_PROPERTIES_TIMOUT = 10

    RETRY_TIMEOUT = 90

    RESET_TIMEOUT = 90

    UDP_SEND_PORT = 5001
    UDP_RECV_PORT = 5002

    @classmethod
    def authorise(cls, device_mac: str) -> str:
        '''Returns the authorise message.'''

        request = {
            "request": {
                "type": "auth_app",
                "params": {
                    "appId": APP_ID,
                    "appToken": APP_TOKEN,
                    "deviceMac": device_mac
                }
            }
        }

        return json.dumps(request)

    @classmethod
    def discover(cls) -> str:
        '''Return the discover message to be broadcast on LAN.'''
        return 'ZIMI'

    @classmethod
    def get(cls, target: str) -> str:
        '''Returns the get messages assocated with a target'''
        request = {"request": {
            "path": "api/v1/controlpoint/", "method": "GET"}}
        if target in ('actions', 'properties', 'states'):
            request["request"]["path"] += target
            return json.dumps(request)
        else:
            raise KeyError(
                'Only "actions", "properties" and "states" are supported')

    @classmethod
    def set(cls, identifier: str, action: str, params: object) -> str:
        '''Returns the set messages assocated with an action'''
        request = {"request": {
            "path": "api/v1/controlpoint/actions", "method": "POST"}}
        if action in ("TurnOn", "TurnOff", "OpenDoor", "CloseDoor"):
            body = {"actions": [{"id": identifier, "action": action}]}
            request["request"]["body"] = body
        elif action in ("Fade", "OpenToPercentage", "SetBrightness", "SetFanSpeed") and params:
            body = {"actions": [
                {"id": identifier,
                 "action": action,
                 "actionParams": params}]}
            request["request"]["body"] = body
        else:
            raise KeyError('Combination of input parameters are not supported')
        return json.dumps(request)

    @classmethod
    def start(cls, device_mac: str, access_token: str) -> str:
        '''Returns the start session message.'''

        request = {
            "request": {
                "type": "start_session",
                "params": {
                    "appId": APP_ID,
                    "deviceMac": device_mac,
                    "accessToken": access_token
                }
            }
        }

        return json.dumps(request)

    @classmethod
    def subscribe(cls) -> str:
        '''Returns the message assocated with state subscribe'''
        request = {"request": {
            "path": "api/v1/subscribe/controlpoint/states", "method": "POST"}}
        return json.dumps(request)

    @classmethod
    def unsubscribe(cls) -> str:
        '''Returns the message assocated with state subscribe'''
        request = {"request": {
            "path": "api/v1/unsubscribe/controlpoint/states", "method": "POST"}}
        return json.dumps(request)
    
    @classmethod
    def inforequest(cls) -> str:
        '''Returns the message assocated with gateway properties request'''
        request = {"request": {
            "path": "api/v1/gateway/properties", "method": "GET"}}
        return json.dumps(request)



#!/usr/bin/env python3

''' Command line parsing for zcc'''

import asyncio
import argparse
import sys


from zcc import ControlPoint, ControlPointDescription, ControlPointDiscoveryService, ControlPointError


def __options(args):

    parser = argparse.ArgumentParser(prog='zcc',
                                     description='''
                                     A command line interface to a ZIMI home controller.

                                     Operates in two basic modes: (a) discovery mode where
                                     the local network is scanned for a ZCC and the details
                                     are printed, or (b) an execute mode where an action is
                                     sent to a device''')

    parser.add_argument("--verbosity", action="store", type=int, default=0,
                        help="verbosity level between 0 and 2")

    command_group = parser.add_argument_group('command')

    cxg = command_group.add_mutually_exclusive_group(required=True)
    cxg.add_argument("--discover", action="store_true",
                     help="discover ZCC and print devices")
    cxg.add_argument("--execute", action="store_true",
                     help="execute an action on a device")

    host_group = parser.add_argument_group('host')
    host_group.add_argument("--host", action="store",
                            help="zcc host name|address")
    host_group.add_argument("--port", action="store", type=int,
                            help="zcc port")
    host_group.add_argument("--timeout", action="store", type=int, default=3,
                            help="zcc timeout value")

    device_group = parser.add_argument_group('device')
    device_group.add_argument("--device", action="store",
                              help="device identifier")
    device_group.add_argument("--action", action="store", type=str,
                              choices=["CloseDoor", "OpenDoor", "TurnOn", "TurnOff",
                                       "OpenToPercentage", "SetBrightness", "SetFanSpeed"])
    device_group.add_argument("--value", action="store",
                              help="device action value (for actions that require)")

    return parser.parse_args(args)


async def main(args):
    '''Main function.'''

    options = __options(args)

    if options.host and options.port:
        description = ControlPointDescription(
            host=options.host, port=options.port)
    else:
        description = await ControlPointDiscoveryService(
            verbosity=options.verbosity).discover()

    controller = ControlPoint(
        description=description, verbosity=options.verbosity)

    await controller.connect()

    if options.discover:

        controller.print_description()

    if options.execute:

        if options.action and options.device:

            device = controller.devices.get(options.device, None)

            if device:
                if 'CloseDoor' in options.action:
                    await device.close_door()
                elif 'OpenDoor' in options.action:
                    await device.open_door()
                elif 'TurnOn' in options.action:
                    await device.turn_on()
                elif 'TurnOff' in options.action:
                    await device.turn_off()
                elif 'OpenToPercentage' in options.action:
                    if options.value:
                        await device.open_to_percentage(options.value)
                    else:
                        raise ControlPointError(
                            'OpenToPercentage requires --value')
                elif 'SetBrightness' in options.action:
                    if options.value:
                        await device.set_brightness(options.value)
                    else:
                        raise ControlPointError(
                            'SetBrightness requires --value')
                elif 'SetFanSpeed' in options.action:
                    if options.value:
                        await device.set_fanspeed(options.value)
                    else:
                        raise ControlPointError(
                            'SetFanSpeed requires --value')
                else:
                    raise ControlPointError(
                        'No valid --device and --action')
            else:
                raise ControlPointError(f'No device {options.device} exists')
        else:
            raise ControlPointError(
                '--execute needs --device and --action options')

    if controller:
        controller.disconnect()


if __name__ == '__main__':

    try:
        asyncio.run(main(sys.argv[1:]))
    except ControlPointError as error:
        print(f"zcc {' '.join(sys.argv[1:])}: {error}")
